/***************************************************************************
                            lvideomodeselector.cpp
                             -------------------
    begin                : Thu Nov 11 1999
    copyright            : (C) 2000 SOT Finnish Software Engineering Ltd.
    author		           : Anton Safonov (Maddy)
    author email         : as@bestlinux.net
    url			             : http://www.bestlinux.net
    email		             : linux@bestlinux.net

	       	      Best Linux - Linux from Finland.
                            http://bestlinux.net

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "lvideomodeselector.h"

#include <qlayout.h>
#include <qlistview.h>
#include <qlabel.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qpushbutton.h>
#include <qbuttongroup.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qfile.h>
#include <qfileinfo.h>

#include "lmonitor.h"
#include "lvideocard.h"
#include "linlinehelp.h"

#include "letcfiles.h"

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

// internal list view item to display the appropriate video modes
class LVideoMode: public QListViewItem
{
public:
    LVideoMode( QListView * parent, LMonitor * monitor, LVideoCard * card,
		QSize res, int refresh, float horizSync,
		float modeClock, const char * modeLine );

    QString text( int ) const;
    QString key( int, bool ) const;
    void setup();

    int depth() const;
    QString depthString( int ) const;

    enum Status{ Ok,
		 BadModeClock, BadHorizSync, BadRefresh,
		 NotEnoughRam, ResTooHigh };
    Status status() const;

    QString reason( Status ) const;

    static bool showAllModes;

    QString modeLine() const;
    QString modeName() const;

    LVideoCard * videoCard() { return vc; }
    LMonitor * monitor() { return m; }

    QSize size() { return r; }

private:
    LMonitor * m;
    LVideoCard * vc;
    QSize r;
    int vs;
    float hs;
    float mc;
    const char * modeline;

    static QString okMessage;
    static QString badModeClockMessage;
    static QString badHorizSyncMessage;
    static QString badRefreshMessage;
    static QString notEnoughRamMessage;
    static QString resTooHighMessage;
    static QString depth8Message;
    static QString depth16Message;
    static QString depth24Message;
    static QString depth32Message;
};


QString LVideoMode::okMessage;
QString LVideoMode::badModeClockMessage;
QString LVideoMode::badHorizSyncMessage;
QString LVideoMode::badRefreshMessage;
QString LVideoMode::notEnoughRamMessage;
QString LVideoMode::resTooHighMessage;

bool LVideoMode::showAllModes;

QString LVideoMode::depth8Message;
QString LVideoMode::depth16Message;
QString LVideoMode::depth24Message;
QString LVideoMode::depth32Message;


LVideoMode::LVideoMode( QListView * parent,
			LMonitor * monitor, LVideoCard * card,
			QSize res, int refresh, float horizSync,
			float modeClock, const char * modeLine )
    : QListViewItem( parent ),
      m( monitor ), vc( card ),
      r( res ), vs( refresh ),
      hs( horizSync ), mc( modeClock ),
      modeline( modeLine )
{
    if ( !okMessage.length() ) {
	okMessage = LVideoModeSelector::tr( "Ok" );
	badModeClockMessage = LVideoModeSelector::tr( "The video card does not offer any compatible clock frequency" );
	badHorizSyncMessage = LVideoModeSelector::tr( "The monitor does not support this horizontal sync frequency" );
	badRefreshMessage = LVideoModeSelector::tr( "The monitor does not support this vertical refresh frequency" );
	notEnoughRamMessage = LVideoModeSelector::tr( "The video card does not have enough RAM" );
	resTooHighMessage = LVideoModeSelector::tr( "Resolution too high for the video card" );

	depth8Message = LVideoModeSelector::tr( "8 bpp" );
	depth16Message = LVideoModeSelector::tr( "8-16 bpp" );
	depth24Message = LVideoModeSelector::tr( "8-24 bpp" );
	depth32Message = LVideoModeSelector::tr( "8-32 bpp" );
    }
}


QString LVideoMode::reason( Status s ) const
{
    if ( s == Ok )
	return okMessage;
    else if ( s == BadModeClock )
	return badModeClockMessage;
    else if ( s == BadHorizSync )
	return badHorizSyncMessage;
    else if ( s == BadRefresh )
	return badRefreshMessage;
    else if ( s == NotEnoughRam )
	return notEnoughRamMessage;
    else if ( s == ResTooHigh )
	return resTooHighMessage;
    return QString::null;
}


int LVideoMode::depth() const
{
    int pixels = r.width() * r.height();
    int ram = vc->memory();
    if ( pixels * 4 <= ram )
	return 32;
    else if ( pixels * 3 <= ram )
	return 24;
    else if ( pixels * 2 <= ram )
	return 16;
    else if ( pixels <= ram )
	return 8;

    return 0;
}


QString LVideoMode::depthString( int d ) const
{
  if ( d >= 32 )
    return depth32Message;
  else if ( d >= 24 )
    return depth24Message;
  else if ( d >= 16 )
    return depth16Message;
  else if ( d >= 8 )
    return depth8Message;
  return QString( "-" );
}


QString LVideoMode::text( int column ) const
{
    QString tmp;
    switch( column ) {
    case 0:
	tmp.sprintf( "%d*%d", r.width(), r.height() );
	break;
    case 1:
	tmp.sprintf( "%d Hz%s", QABS( vs ), vs > 0 ? "" : " interl." );
	break;
    case 2:
	tmp = depthString( depth() );
	break;
    case 3:
	tmp = showAllModes ? reason( status() ) : okMessage;
	break;
    case 4:
	tmp.sprintf( "%g kHz", hs );
	break;
    case 5:
	tmp.sprintf( "%g MHz", mc );
	break;
    }
    return tmp;
}


QString LVideoMode::key( int column, bool ) const
{
    QString tmp;
    switch( column ) {
    case 0:
	tmp.sprintf( "%04d%04d", r.width(), r.height() );
	break;
    case 1:
	tmp.sprintf( "%04d", vs > 0 ? vs : vs/-2 );
	break;
    case 2:
	tmp.sprintf( "%02d", depth() );
	break;
    case 3:
	tmp = reason( status() );
	break;
    case 4:
	tmp.sprintf( "%4.3f", hs );
	break;
    case 5:
	tmp.sprintf( "%7.3f", mc );
	break;
    case 6:
	tmp.sprintf( "%04d %04d %04d %04d",
		     r.width(), r.height(),
		     vs > 0 ? vs : vs/-2, depth() );
	break;
    }
    return tmp;
}


void LVideoMode::setup()
{
    QListViewItem::setup();
    bool ok = status() == Ok;
    setHeight( ok || showAllModes ? 18 : 0 ); // ### 18 is a hack
    setSelectable( ok );
}


LVideoMode::Status LVideoMode::status() const
{
    if ( !m->supportsHorizontalFrequency( hs ) )
	return BadHorizSync;
    else if ( !m->supportsVerticalFrequency( vs ) )
	return BadRefresh;
    else if ( !vc->supportsModeClock( mc ) )
	return BadModeClock;
    else if ( r.width()*r.height() > vc->memory() )
	return NotEnoughRam;
    else if ( vc->maxResolution().width() > 0 &&
	      ( r.width() > vc->maxResolution().width() ||
		r.height() > vc->maxResolution().height() ) )
	return ResTooHigh;

    return Ok;
}


QString LVideoMode::modeLine() const
{
    return modeline;
}


QString LVideoMode::modeName() const
{
    QString tmp;
    tmp.sprintf( "\"%dx%d/%dHz\"", r.width(), r.height(), vs );
    return tmp;
}


LVideoModeSelector::LVideoModeSelector( LMonitor * monitor, LVideoCard * card,
					LMouse * mouse, LKeyboard * keyboard,
					QWidget * parent, const char * name )
    : QWidget( parent, name ), m( monitor ), vc( card ),
      mouseHack (mouse ), keyboardHack( keyboard )
{
    QHBoxLayout * tll = new QHBoxLayout( this, 0, 6 );

    QVBoxLayout * l = new QVBoxLayout( 6 );
		tll->addStretch( 1 );
    tll->addLayout( l, 8 );
		tll->addStretch( 1 );		

		l->addSpacing( 10 );
    modes = new QListView( this );
		QFont font( QFont::defaultFont() );
		font.setFamily( "times" );
		font.setPointSize( 10 );
    modes->setFont( font );
    l->addWidget( modes );
    modes->setAllColumnsShowFocus( TRUE );
    modes->setFrameStyle( QFrame::WinPanel + QFrame::Sunken );
    modes->setItemMargin( 3 );
    QString header;
    QFontMetrics fm( modes->fontMetrics() );
    header = tr("Resolution");
    modes->addColumn( header, fm.width( header ) + 8 );
    header = tr("Refresh");
    modes->addColumn( header, fm.width( header ) + 8 );
    header = tr("Depth");
    modes->addColumn( header, fm.width( header ) + 8 );
    header = tr("Availability");
    modes->addColumn( header, fm.width( header ) + 8 );
    header = tr("Horiz. Sync");
    modes->addColumn( header, fm.width( header ) + 8 );
    header = tr("Mode Clock");
    modes->addColumn( header, fm.width( header ) + 8 );
    modes->setColumnWidthMode( 0, QListView::Maximum );
    modes->setColumnWidthMode( 1, QListView::Maximum );
    modes->setColumnWidthMode( 2, QListView::Maximum );
    modes->setColumnWidthMode( 3, QListView::Maximum );
    modes->setColumnWidthMode( 4, QListView::Maximum );
    modes->setColumnWidthMode( 5, QListView::Maximum );

    // the following numbers are from XFree86.
    // $XConsortium: mondata.tcl /main/4 1996/10/28 05:42:19 kaleb $
    // $XFree86: xc/programs/Xserver/hw/xfree86/XF86Setup/mondata.tcl,v 3.8.2.4 1998/02/15 16:08:53 hohndel Exp $
    add( QSize( 640, 400 ), 60, 24, 16.442,
	 "16.442 640 649 656 663 400 407 415 440" );
    add( QSize( 640, 400 ), 70, 31.5, 25.175,
	 "25.175 640  664  760  800   400  409  411  450" );
    add( QSize( 640, 480 ), 60, 31.5, 25.175,
	 "25.175 640  664  760  800   480  491  493  525" );
    add( QSize( 800, 600 ), 56, 35.15, 36,
	 "36     800  824  896 1024   600  601  603  625" );
    add( QSize( 1024, 768 ), -87, 35.5, 44.9,
	 "44.9  1024 1048 1208 1264   768  776  784  817 Interlace" );
    add( QSize( 640, 480 ), 67, 35.0, 28,
	 "28     640  664  760  800   480  491  493  525" );
    add( QSize( 640, 400 ), 85, 37.86, 31.5,
	 "31.5    640  672 736   832   400  401  404  445 -HSync +VSync" );
    add( QSize( 640, 480 ), 70, 36.5, 31.5,
	 "31.5   640  680  720  864   480  488  491  521" );
    add( QSize( 640, 480 ), 75, 37.50, 31.5,
	 "31.5   640  656  720  840   480  481  484  500 -HSync -VSync" );
    add( QSize( 800, 600 ), 60, 37.8, 40,
	 "40     800  840  968 1056   600  601  605  628 +HSync +VSync" );
    add( QSize( 640, 480 ), 85, 43.27, 36,
	 "36     640  696  752  832   480  481  484  509 -HSync -VSync" );
    add( QSize( 1152, 864 ), -89, 44, 65,
	 "65    1152 1168 1384 1480   864  865  875  985 Interlace" );
    add( QSize( 800, 600 ), 72, 48.0, 50,
	 "50     800  856  976 1040   600  637  643  666 +HSync +VSync" );
    add( QSize( 1024, 768 ), 60, 48.4, 65,
	 "65    1024 1032 1176 1344   768  771  777  806 -HSync -VSync" );
    add( QSize( 640, 480 ), 100, 53.01, 45.8,
	 "45.8   640  672  768  864   480  488  494  530 -HSync -VSync" );
    add( QSize( 1152, 864 ), 60, 53.5, 89.9,
	 "89.9  1152 1216 1472 1680   864  868  876  892 -HSync -VSync" );
    add( QSize( 800, 600 ), 85, 55.84, 60.75,
	 "60.75   800  864  928 1088   600  616  621  657 -HSync -VSync" );
    add( QSize( 1024, 768 ), 70, 56.5, 75,
	 "75    1024 1048 1184 1328   768  771  777  806 -HSync -VSync" );
    add( QSize( 1280, 1024 ), -87, 51, 80,
	 "80    1280 1296 1512 1568  1024 1025 1037 1165 Interlace" );
    add( QSize( 1024, 768 ), 76, 62.5, 85,
	 "85    1024 1032 1152 1360   768  784  787  823" );
    add( QSize( 1152, 864 ), 70, 62.4, 92,
	 "92    1152 1208 1368 1474   864  865  875  895" );
    add( QSize( 800, 600 ), 100, 64.02, 69.65,
	 "69.650 800  864  928 1088   600  604  610  640 -HSync -VSync" );
    add( QSize( 1024, 768 ), 85, 70.24, 98.9,
	 "98.9  1024 1056 1216 1408   768 782 788 822 -HSync -VSync" );
    add( QSize( 1152, 864 ), 78, 70.8, 110,
	 "110    1152 1240 1324 1552   864  864  876  908" );
    add( QSize( 1152, 864 ), 84, 76.0, 135,
	 "135    1152 1464 1592 1776   864  864  876  908" );
    add( QSize( 1280, 1024 ), 61, 64.2, 110,
	 "110    1280 1328 1512 1712  1024 1025 1028 1054" );
    add( QSize( 1280, 1024 ), 70, 74.59, 126.5,
	 "126.5  1280 1312 1472 1696  1024 1032 1040 1068 -HSync -VSync" );
    add( QSize( 1600, 1200 ), 60, 75.00, 162,
	 "162    1600 1664 1856 2160  1200 1201 1204 1250 +HSync +VSync" );
    add( QSize( 1280, 1024 ), 74, 78.85, 135,
	 "135    1280 1312 1456 1712  1024 1027 1030 1064" );
    add( QSize( 1024, 768 ), 100, 80.21, 115.5,
	 "115.5  1024 1056 1248 1440  768  771  781  802 -HSync -VSync" );
    add( QSize( 1280, 1024 ), 76, 81.13, 135,
	 "135    1280 1312 1416 1664  1024 1027 1030 1064" );
    add( QSize( 1600, 1200 ), 70, 87.50, 189,
	 "189    1600 1664 1856 2160  1200 1201 1204 1250 -HSync -VSync" );
    add( QSize( 1152, 864 ), 100, 89.62, 137.65,
	 "137.65 1152 1184 1312 1536   864  866  885  902 -HSync -VSync" );
    add( QSize( 1280, 1024 ), 85, 91.15, 157.5,
	 "157.5  1280 1344 1504 1728  1024 1025 1028 1072 +HSync +VSync" );
    add( QSize( 1600, 1200 ), 75, 93.75, 202.5,
	 "202.5  1600 1664 1856 2160  1200 1201 1204 1250 +HSync +VSync" );
    add( QSize( 1600, 1200 ), 85, 105.77, 220,
	 "220    1600 1616 1808 2080  1200 1204 1207 1244 +HSync +VSync" );
    add( QSize( 1280, 1024 ), 100, 107.16, 181.75,
	 "181.75 1280 1312 1440 1696  1024 1031 1046 1072 -HSync -VSync" );
    add( QSize( 1800, 1440 ), 64, 96.15, 230,
	 "230    1800 1896 2088 2392 1440 1441 1444 1490 +HSync +VSync" );
    add( QSize( 1800, 1440 ), 70, 104.52, 250,
	 "250    1800 1896 2088 2392 1440 1441 1444 1490 +HSync +VSync" );
    add( QSize( 320, 200 ), 70, 31.5, 12.588,
	 "12.588 320  336  384  400   200  204  205  225 Doublescan" );
    add( QSize( 320, 240 ), 60, 31.5, 12.588,
	 "12.588 320  336  384  400   240  245  246  262 Doublescan" );
    add( QSize( 320, 240 ), 72, 36.5, 15.750,
	 "15.750 320  336  384  400   240  244  246  262 Doublescan" );
    add( QSize( 400, 300 ), 56, 35.2, 18,
	 "18     400  416  448  512   300  301  302  312 Doublescan" );
    add( QSize( 400, 300 ), 60, 37.8, 20,
	 "20     400  416  480  528   300  301  303  314 Doublescan" );
    add( QSize( 400, 300 ), 72, 48.0, 25,
	 "25     400  424  488  520   300  319  322  333 Doublescan" );
    add( QSize( 480, 300 ), 56, 35.2, 21.656,
	 "21.656 480  496  536  616   300  301  302  312 Doublescan" );
    add( QSize( 480, 300 ), 60, 37.8, 23.890,
	 "23.890 480  496  576  632   300  301  303  314 Doublescan" );
    add( QSize( 480, 300 ), 63, 39.6, 25,
	 "25     480  496  576  632   300  301  303  314 Doublescan" );
    add( QSize( 480, 300 ), 72, 48.0, 29.952,
	 "29.952 480  504  584  624   300  319  322  333 Doublescan" );
    add( QSize( 512, 384 ), 78, 31.50, 20.16,
	 "20.160 512  528  592  640   384  385  388  404 -HSync -VSync" );
    add( QSize( 512, 384 ), 85, 34.38, 22,
	 "22     512  528  592  640   384  385  388  404 -HSync -VSync" );

    // magic sorting column
    modes->setSorting( 6, FALSE );
    modes->setSelected( modes->firstChild(), FALSE );

    connect( modes, SIGNAL(selectionChanged()),
	     this, SLOT(updateDepthBoxes()) );
    
    QGridLayout * buttons = new QGridLayout( 1, 1, 6 );
    l->addLayout( buttons );

    colDepth = new QComboBox(this);
    buttons->addWidget(colDepth,0,0);
    connect(colDepth, SIGNAL(activated(int)), this, SLOT(updateBoxes(int)));

    virtRes = new QComboBox(this);
    buttons->addWidget(virtRes,1,0);
    connect(virtRes, SIGNAL(activated(int)), this, SLOT(updateBoxes(int)));

    allModes = new QCheckBox(tr("Show &all modes"), this);
    buttons->addWidget(allModes, 2,0);

    tryModeButton = new QPushButton( tr( "&Test this mode" ), this);
    connect( tryModeButton, SIGNAL(clicked()),
	     this, SLOT(tryMode()) );
    QVBoxLayout * tryModeLayout = new QVBoxLayout( 6 );
    buttons->addLayout( tryModeLayout, 0, 1 );
    tryModeLayout->addWidget( tryModeButton, 0, AlignRight + AlignTop );
    tryModeLayout->addSpacing( 6 );

/*    LInlineHelp* inlineHelp = new LInlineHelp( this );
    inlineHelp->setDocument( "videomode_overview.qml" );
    tll->addWidget( inlineHelp, 0 );*/
		l->addSpacing( 10 );

    connect( allModes, SIGNAL(toggled(bool)),
	     this, SLOT(setShowAllModes(bool)) );

    // set up the virtual resolutions
    vResolutions.append(new QSize(800,600));
    vResolutions.append(new QSize(1024,768));
    vResolutions.append(new QSize(1152,864));
    vResolutions.append(new QSize(1280,1024));
    vResolutions.append(new QSize(1600,1200));
    vResolutions.append(new QSize(1800,1440));

    oldCol = oldVirt = -1;
}


void LVideoModeSelector::add( QSize res, int refresh,
		      float horizSync, float modeClock, const char * modeLine )
{
    (void)new LVideoMode( modes, m, vc, res, refresh,
			  horizSync, modeClock, modeLine );
}


void LVideoModeSelector::show()
{
    setShowAllModes( LVideoMode::showAllModes );
    QWidget::show();
}


void LVideoModeSelector::setShowAllModes( bool enable )
{
    LVideoMode::showAllModes = enable;
    QListViewItem * i = modes->firstChild();
    while( i ) {
			i->setup();
			i = i->nextSibling();
    }

    i = modes->selectedItem();
    if ( !i || i->height() == 0 )
			i = modes->firstChild();
    while( i && i->height() == 0 )
			i = i->nextSibling();

    if ( i ) {
			modes->setSelected( i, TRUE );
			modes->ensureItemVisible( i );
    }
}


void LVideoModeSelector::updateBoxes(int)
{
  updateDepthBoxes();
}


void LVideoModeSelector::updateDepthBoxes()
{
  // the color depths
  LVideoMode * m = (LVideoMode*)(modes->selectedItem());
  
  int d = 0;
  if ( m )
    d = m->depth();
  int old = colDepth->currentItem();
  
  colDepth->clear();
  colDepth->insertItem(tr("256 colors (8bpp)"));
  if (d >= 15)
    colDepth->insertItem(tr("Thousands colors (15/16bpp)"));
  if (d >= 24)
    colDepth->insertItem(tr("Millions colors (24bpp)"));
  if (d >=32)
    colDepth->insertItem(tr("True color (32bpp)"));

  if (oldCol == -1)
    {
      oldCol = old;
      colDepth->setCurrentItem(colDepth->count()-1);
    }
  else
    if (old < colDepth->count())
      colDepth->setCurrentItem(old);
    else
      colDepth->setCurrentItem(colDepth->count()-1);
  
  // the virtual resolutions
  int min=0;
  for (unsigned int i=0; i < vResolutions.count()-1; i++)
    {
      QSize *res = vResolutions.at(i);
      if (res->width() <= width() || res->height() <= height())
	min = i+1;
    }

  old = virtRes->currentItem();
  virtRes->clear();
  resEntries.clear();
  virtRes->insertItem(tr("no virtual desktop"));
  resEntries.append(new QSize(width(),height()));
  for (unsigned int i=min; i < vResolutions.count(); i++)
    {
      int pix = vResolutions.at(i)->width() * vResolutions.at(i)->height();
      if (pix * defaultColorDepth()/8 <= vc->memory())
	{
 	  QString item(tr("%0 x %1 (virtual)"));
	  QSize *res = vResolutions.at(i);
	  virtRes->insertItem(item.arg(res->width()).arg(res->height()));
	  resEntries.append(new QSize(*res));	  
	}
    }

  if (oldVirt == -1)
    {
      oldVirt = old;
      if (m && m->size().width() < 800 && m->size().height() < 600 && virtRes->count() > 0)
	virtRes->setCurrentItem(virtRes->count()-1);
    }
  else
    if (old < virtRes->count())
      virtRes->setCurrentItem(old);
}


void LVideoModeSelector::tryMode()
{
    LVideoMode * m = (LVideoMode*)(modes->selectedItem());
    if ( !m )
			return;

    if ( QMessageBox::information( this, tr( "Test Video Mode" ),
				   tr( "The screen will now switch to the selected mode.\n\n"
				       "It will switch back when you click the left mouse\n"
				       "button, or automatically after ten seconds.\n"
				       "\nNote that testing may hang your system.\nIf it hangs,"
				       "you will have to repeat the installation\nwithout testing."),
				   tr( "Ok" ), tr( "Abort" ), QString::null,
				   0, 1 ) == 1 )
			return;

		::mkdir( "/tmp/X11", 0775 );
    LEtcFiles etc( "/tmp" );
    etc.writeXConfig( m->videoCard(), m->monitor(), this,
		      mouseHack, keyboardHack, "/dev/mouse"/*"/tmp/mouselnk"*/ );

    QApplication::flushX();
    QString result;
    int pid = ::fork();
    if ( pid == 0 ) {
			int i = getdtablesize();
			while( --i >= 0 )
	  	  ::close( i );
			::open( "/dev/null", O_RDWR );
			::dup( 0 );
			::dup( 1 );
	
			execl( "./ltestvideomode", "./ltestvideomode",
	       "Press any button or key to finish testing,\nor wait ten seconds.",
	       0 );
			exit( 0 );
    }
printf( "lvideomodeselector::tryMode pid: %d\n", pid );
    ::waitpid(pid, NULL, 0);
}


QString LVideoModeSelector::line() const
{
    LVideoMode * m = (LVideoMode*)(modes->selectedItem());
    if ( m )
			return m->modeLine();
    return QString ( "25.175 640  664  760  800   480  491  493  525" );
}


QString LVideoModeSelector::name() const
{
    LVideoMode * m = (LVideoMode*)(modes->selectedItem());
    if ( m )
			return m->modeName();
    return QString ( "\"640x480/60Hz\"" );
}


int LVideoModeSelector::defaultColorDepth()
{
  switch (colDepth->currentItem())
    {
    case 0:
      return 8;
    case 1:
      return 16;
    case 2:
      return 24;
    case 3:
      return 32;
    default:
      return 8;
    }
}


int LVideoModeSelector::width()
{
    LVideoMode * m = (LVideoMode*)(modes->selectedItem());
    if ( m )
	return m->size().width();
    return 640;
}


int LVideoModeSelector::height()
{
    LVideoMode * m = (LVideoMode*)(modes->selectedItem());
    if ( m )
	return m->size().height();
    return 480;
}


QSize LVideoModeSelector::virtualResolution()
{
  unsigned int c = virtRes->currentItem();

  if (c < resEntries.count())
    return *resEntries.at(c);
  else
    return QSize(width(), height());
}













