#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/time.h>
#include <dirent.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "mdetect.h"

static void		usage(int);
static struct device *	stage1(struct device *);
static void		stage2(struct device *);
static int		handle_device(struct device *, fd_set *, unsigned long);
static struct device *	open_devices(void);
static struct device *	open_device(const char *name);

static int		opt_verbose = 0;
static int		opt_repeater = 0;

static fd_set		mouse_set;

int
main(int argc, char **argv)
{
	struct device	*dev, *tmp;
	int		c;

	while ((c = getopt(argc, argv, "hrv")) != -1) {
		switch (c) {
		case 'v':
			opt_verbose = 1;
			break;
		case 'r':
			opt_repeater = 1;
			break;
		case 'h':
			usage(0);
		default:
			usage(1);
		}
	}

	if (opt_verbose)
		printf("The mouse police never sleeps.\n");
	if (!(dev = open_devices()))
		return 1;

	if (opt_verbose) {
		printf("Found the following devices:\n");
		for (tmp = dev; tmp; tmp = tmp->next)
			printf("   %s\n", tmp->name);
	}

	/* Now listen on all devices */
	dev = stage1(dev);

	if (dev == NULL)
		goto give_up;

	stage2(dev);
	return 0;

give_up:
	while (1) pause();
	return 1;
}

static void
usage(int exitval)
{
	fprintf(stderr, "Usage: mdetect [-hrv]\n");
	exit(exitval);
}

static struct device *
stage1(struct device *dev)
{
	struct device	*tmp;
	unsigned long	timeout;
	time_t		start_time;
	int		max = -1;

	FD_ZERO(&mouse_set);
	for (tmp = dev; tmp; tmp = tmp->next) {
		FD_SET(tmp->fd, &mouse_set);
		if (max < tmp->fd)
			max = tmp->fd;
	}

	timeout = 0;
	time(&start_time);
	while (1) {
		struct timeval	tv, now, then, *tvp = NULL;
		unsigned long	elapsed;
		fd_set		set;

		memcpy(&set, &mouse_set, sizeof(mouse_set));
		if (timeout) {
			tv.tv_sec  = 0;
			tv.tv_usec = timeout * 1000;
			tvp = &tv;
		}
		gettimeofday(&now, NULL);
		if (select(max + 1, &set, NULL, NULL, tvp) < 0) {
			perror("select");
			exit(1);
		}
		gettimeofday(&then, NULL);
		elapsed = (then.tv_sec  - now.tv_sec)  * 1000
			+ (then.tv_usec - now.tv_usec) / 1000;

		timeout = 0;
		for (tmp = dev; tmp; tmp = tmp->next) {
			handle_device(tmp, &set, elapsed);
			if (tmp->timeout > timeout)
				timeout = tmp->timeout;
			if (tmp->driver->good > 20)
				return tmp;
		}
	}

	return NULL;
}

static void
stage2(struct device *dev)
{
	struct mouse	*driver = dev->driver;
	const char	*name;
	unsigned char	c;

	name = driver->name;
	if (driver->report)
		name = driver->report(driver);
	printf("\nDetected %s mouse on %s\n", driver->name, dev->name);
	fflush(stdout);

	/* Repeater code not yet complete */
	if (opt_repeater) {
		while (1) {
			if (read(dev->fd, &c, 1) >= 0)
				driver->packet(dev, driver, c);
		}
	}
}

static int
handle_device(struct device *dev, fd_set *set, unsigned long elapsed)
{
	struct mouse	*driver = dev->driver;

	if (dev->fd >= 0 && FD_ISSET(dev->fd, set)) {
		unsigned char	c;

		dev->timeout = 0;
		if (read(dev->fd, &c, 1) < 0)
			goto bad;
		if (driver->packet(dev, dev->driver, c) < 0)
			goto bad;
	} else if (dev->timeout) {
		if (elapsed > dev->timeout) {
			if (!driver->timeout || driver->timeout(dev) < 0)
				goto bad;
		}
		dev->timeout = 0;
	}

	return 0;

bad:
	if (opt_verbose)
		printf("\nClosing garbage device %s.\n", dev->name);
	FD_CLR(dev->fd, &mouse_set);
	close(dev->fd);
	dev->fd = -1;
	dev->timeout = 0;
	return -1;
}

/*
 * Send an event
 */
int
send_event(struct mouse *driver)
{
	static struct mouse	*last = NULL;
	int			ddx, ddy;

	if (last != driver && opt_verbose)
		printf("\n");
	last = driver;

	if ((ddx = driver->lx - driver->dx) < 0)
		ddx = -ddx;
	if ((ddy = driver->ly - driver->dy) < 0)
		ddy = -ddy;
	if (opt_verbose) {
		unsigned int	b =driver->buttons;

		printf("%-12s %c%c%c %4d %4d (der %3d) %2d%%   \r",
			driver->name,
			(b & BUTTON_LEFT)?   'L' : ' ',
			(b & BUTTON_MIDDLE)? 'M' : ' ',
			(b & BUTTON_RIGHT)?  'R' : ' ',
			driver->dx, driver->dy,
			ddx + ddy,
			100 * driver->good / 20);
		fflush(stdout);
	}

	/* Consistency checks for the first 20 events.
	 * Note that the button check is the most effective one.
	 * Motion values aren't that far off most of the time,
	 * but buttons frequently are.
	 * XXX: Tue Jun 22 11:12:25 MET DST 1999
	 *  I disabled motion checks. All they effected was requiring
	 *  the user to move the mouse `slowly' according to various
	 *  reports.
	 *
	 *  Without motion checks, mdetect still seems to work reliably,
	 *  plus you can jerk the mouse as much as you please.
	 *			--okir
	 */
	if (driver->good < 20) {
		if ((driver->buttons & (BUTTON_MIDDLE|BUTTON_RIGHT))
//#if 0
		 || ddx > 25 //30
		 || (driver->dx < -200 || driver->dx > 200)
		 || (driver->dy < -200 || driver->dy > 200)
//#endif
		) {
			if (opt_verbose)
				printf("\n");
			driver->bad++;
			return -1;
		}
	} else {
		/* Deliver the event to the fifo */
	}

	driver->good++;
	return 0;
}

/* Open all devices */
static struct device *
open_devices(void)
{
	struct device	*dev = NULL;
	DIR		*dir;
	struct dirent	*de;

	if (!(dir = opendir("/dev"))) {
		perror("/dev");
		return NULL;
	}

	while ((de = readdir(dir)) != 0) {
		struct device	*tmp;

		if ((tmp = open_device(de->d_name)) != 0) {
			tmp->next = dev;
			dev = tmp;
		}
	}

	return dev;
}

static struct device *
open_device(const char *name)
{
	struct device	temp, *dev;
	struct mouse	*driver;
	struct stat	stb;
	int		serial = 0;

	memset(&temp, 0, sizeof(temp));
	strcat(strcpy(temp.name, "/dev/"), name);

	if (!strncmp(name, "ttyS", 4)) {
		driver = get_mouse("serial");
		serial = 1;
	} else {
		/* Not a serial device, try bus mice */
		if (lstat(temp.name, &stb) < 0)
			return NULL;
		if (!S_ISCHR(stb.st_mode) || major(stb.st_rdev) != 10)
			return NULL;

		if (!(driver = get_mouse(name)))
			return NULL;
	}
	temp.driver = driver;
	if ((temp.fd = open(temp.name, O_RDWR)) < 0)
		return NULL;
	dev = (struct device *) malloc(sizeof(*dev));
	memcpy(dev, &temp, sizeof(*dev));

	if (driver->init) {
		if (dev->driver->init(dev, driver) < 0) {
			close(dev->fd);
			free(dev);
			return NULL;
		}
	}
	return dev;
}
